#! python3  # noqa: E265

"""
    Cross-platform shortcuts manager. Derived from pycrosskit, by  Jiri Otoupal.
    See also: https://github.com/newville/pyshortcuts/
    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################
# standard
import logging
import os
import re
import stat
import sys
from collections.abc import Iterable
from pathlib import Path
from string import Template
from sys import platform as opersys

# Imports depending on operating system
if opersys == "win32":
    """windows"""
    import pythoncom
    import win32com.client
    from win32comext.shell import shell, shellcon


# package
from qgis_deployment_toolbelt.__about__ import __title__, __version__
from qgis_deployment_toolbelt.constants import (
    QGIS_BIN_WINDOWS_FILENAME,
    QGIS_LTR_BIN_WINDOWS_FILENAME,
    OSConfiguration,
)
from qgis_deployment_toolbelt.utils.check_path import check_path
from qgis_deployment_toolbelt.utils.slugger import sluggy

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class ApplicationShortcut:
    def __init__(
        self,
        name: str,
        exec_path: str | Path,
        exec_arguments: Iterable[str] | None = None,
        description: str | None = None,
        icon_path: str | Path | None = None,
        work_dir: str | Path | None = None,
    ):
        """Initialize a shortcut object.

        Args:
            name (str): name of the shortcut that will be created
            exec_path (str | Path): path to the executable (which should exist)
            exec_arguments (Iterable[str] | None, optional): list of arguments and \
                options to pass to the executable. Defaults to None.
            description (str | None, optional): shortcut description. Defaults to None.
            icon_path (str | Path | None, optional): path to icon file. Defaults to None.
            work_dir (str | Path | None, optional): current folder where to start the \
                executable. Defaults to None. In QDT, it's the profile folder.

        Raises:
            ValueError: if shortcut name contains invalid characters (depending on \
                operating system)
        """
        # retrieve operating system specific configuration
        self.os_config = OSConfiguration.from_opersys()

        # -- CHECK TYPE AND STORE ATTRIBUTES
        # mandatory
        if isinstance(name, str):
            if not self.os_config.valid_shortcut_name(name):
                raise ValueError(f"Shortcut name {name} contains invalid characters.")
            self.name = name
        else:
            raise TypeError(f"Shortcut name must be a string, not {type(name)}.")

        if isinstance(exec_path, (str, Path)):
            self.exec_path = Path(exec_path)
            if not self.exec_path.exists():
                # helper to handle common typo error on executable name on Windows
                if (
                    self.exec_path.name.endswith(QGIS_BIN_WINDOWS_FILENAME)
                    and self.exec_path.with_name(QGIS_LTR_BIN_WINDOWS_FILENAME).exists()
                ):
                    logger.warning(
                        f"Executable set does not exist: {self.exec_path} "
                        f"but {self.exec_path.with_name('qgis-ltr-bin.exe')} does, so "
                        "this one will be used instead. Check and fix your scenario."
                    )
                    self.exec_path = self.exec_path.with_name(
                        QGIS_LTR_BIN_WINDOWS_FILENAME
                    )
                elif (
                    self.exec_path.name.endswith(QGIS_LTR_BIN_WINDOWS_FILENAME)
                    and self.exec_path.with_name(QGIS_BIN_WINDOWS_FILENAME).exists()
                ):
                    logger.warning(
                        f"Executable set does not exist: {self.exec_path} "
                        f"but {self.exec_path.with_name('qgis-bin.exe')} does, so "
                        "this one will be used instead. Check and fix your scenario."
                    )
                    self.exec_path = self.exec_path.with_name(QGIS_BIN_WINDOWS_FILENAME)
                else:
                    logger.warning(
                        f"Executable does not exist: {self.exec_path}. "
                        "Shortcuts might not work. Check and fix your scenario."
                    )
        else:
            raise TypeError(
                f"exec_path must be a string or pathlib.Path, not {type(exec_path)}"
            )

        # optional
        if isinstance(exec_arguments, (tuple, list, type(None))):
            self.exec_arguments = self.check_exec_arguments(exec_arguments)
        else:
            raise TypeError(
                f"If defined, exec_arguments must be a tuple or list, not {type(exec_arguments)}"
            )
        if isinstance(description, (str, type(None))):
            self.description = description
        else:
            raise TypeError(
                f"If defined, description must be a string, not {type(description)}"
            )
        if isinstance(icon_path, (str, Path, type(None))):
            self.icon_path = self.check_icon_path(icon_path)
        else:
            raise TypeError(
                f"If defined, icon_path must be a string or pathlib.Path, not {type(icon_path)}"
            )
        if isinstance(work_dir, (str, Path, type(None))):
            self.work_dir = self.check_work_dir(work_dir)
        else:
            raise TypeError(
                f"If defined, work_dir must be a string or pathlib.Path, not {type(work_dir)}"
            )

    def create(
        self,
        desktop: bool = False,
        start_menu: bool = False,
    ) -> tuple[Path | None, Path | None]:
        """Creates shortcuts.

        Args:
            desktop (bool, optional): True to generate a Desktop shortcut. \
                Defaults to False.
            start_menu (bool, optional): True to generate a 'Start Menu' shortcut. \
                Defaults to False.

        Raises:
            TypeError: if options are not booleans

        Returns:
            tuple[Path, Path]: desktop and startmenu path
        """
        if isinstance(desktop, bool):
            self.desktop = desktop
        else:
            raise TypeError(f"'desktop' option must be a boolean, not {type(desktop)}")
        if isinstance(start_menu, bool):
            self.start_menu = start_menu
        else:
            raise TypeError(
                f"'start_menu' option must be a boolean, not {type(start_menu)}"
            )

        if not desktop and not start_menu:
            logger.debug(
                "Shortcut will not be created because both desktop and start "
                "menu options are False."
            )
            return (None, None)

        if opersys == "win32":
            return self.win32_create()
        elif opersys == "linux":
            return self.freedesktop_create()
        else:
            return (None, None)

    def check_exec_arguments(self, exec_arguments: Iterable[str] | None) -> str | None:
        """Check if exec_arguments are valid.

        Args:
            exec_arguments (Iterable[str] | None): input executable arguments to check

        Returns:
            str | None: str of arguments
        """
        if not exec_arguments:
            return None

        # iterate and separate with spaces
        return " ".join(exec_arguments)

    def check_icon_path(self, icon_path: str | Path | None) -> Path | None:
        """Check icon path and return full path if it exists.

        Args:
            icon_path (str | Path | None): input icon path to check

        Returns:
            Path | None: icon path as Path if str or Path, else None
        """
        if icon_path is None:
            logger.debug(
                f"Shortcut '{self.name}' has no icon specified. Fallback to default "
                "QGIS icon."
            )
            return self.os_config.shortcut_icon_default_path

        # checks
        if check_path(
            input_path=icon_path,
            must_be_a_file=True,
            must_be_readable=True,
            must_exists=True,
            raise_error=False,
        ):
            return Path(icon_path).resolve()
        else:
            logger.warning(f"Icon does not exist: {icon_path}")
            return None

    def check_work_dir(self, work_dir: str | Path | None) -> Path | None:
        """Check work dir and return full path if it exists.

        Args:
            work_dir (str | Path | None): input work dir to check

        Returns:
            Path | None: work dir as Path if str or Path, else None
        """
        if not work_dir:
            return None
        # store as path
        work_dir = Path(work_dir)
        # checks
        if work_dir.is_dir():
            return work_dir.resolve()
        else:
            logger.warning(f"Work folder does not exist: {work_dir}")
            return None

    # -- PROPERTIES --------------------------------------------------------------
    @property
    def desktop_path(self) -> Path:
        """Return the user Desktop folder.

        Partly inspired from https://stackoverflow.com/a/13742626/2556577.

        Returns:
            Path: path to the current user Desktop folder.
        """
        default_value = Path(Path.home(), "Desktop")

        if opersys == "win32":
            return Path(shell.SHGetFolderPath(0, shellcon.CSIDL_DESKTOP, None, 0))
        elif opersys in ("darwin", "linux"):
            # have a look to XDG configuration file with user folders
            config_xdg_user_dirs = Path(self.homedir_path, ".config", "user-dirs.dirs")
            is_config_file_usable = check_path(
                input_path=config_xdg_user_dirs,
                must_be_a_file=True,
                must_be_readable=True,
                must_exists=True,
                raise_error=False,
            )
            if is_config_file_usable:
                with config_xdg_user_dirs.open(mode="r", encoding="UTF-8") as bf_config:
                    data = bf_config.read()

                desktop_paths = re.findall('XDG_DESKTOP_DIR="([^"]*)', data)
                if len(desktop_paths):
                    return Path(
                        re.sub(r"\$HOME", os.path.expanduser("~"), desktop_paths[0])
                    )

            return default_value
        else:
            logger.warning(
                f"Unrecognized operating system ({opersys}) so path to the "
                f"user desktop is using a fallback value: {default_value}"
            )
            return default_value

    @property
    def homedir_path(self) -> Path | None:
        """Return home directory.

        For Windows, note that we return `CSIDL_PROFILE`, not `CSIDL_APPDATA`,
        `CSIDL_LOCAL_APPDATA` or `CSIDL_COMMON_APPDATA`.
        See: https://www.nirsoft.net/articles/find_special_folder_location.html
        TODO: evaluate use of platformdirs

        Returns:
            Path | None: path to the user home
        """
        if opersys == "win32":
            return Path(shell.SHGetFolderPath(0, shellcon.CSIDL_PROFILE, 0, 0))
        elif opersys == "linux":
            home = None
            try:
                home = Path.home()
            except RuntimeError as err:
                logger.debug("Home directory can’t be resolved with pathlib: %s", err)
            # try another way
            if home is None:
                home = os.path.expanduser("~")
            if home is None:
                home = os.path.normpath(os.environ.get("HOME", os.path.abspath(".")))
            return Path(home)
        elif opersys == "darwin":
            return Path(os.path.expanduser("~"))
        else:
            logger.error(f"Unrecognized operating system: {opersys}.")
            return None

    @property
    def startmenu_path(self) -> Path | None:
        """Return user Start Menu Programs folder.

        For Windows, note that we return `CSIDL_PROGRAMS` not `CSIDL_COMMON_PROGRAMS`.

        Returns:
            Path: path to the Start Menu Programs folder
        """
        if opersys == "win32":
            return Path(shell.SHGetFolderPath(0, shellcon.CSIDL_PROGRAMS, None, 0))
        elif opersys in ("darwin", "linux"):
            if isinstance(self.homedir_path, Path):
                return self.homedir_path / ".local/share/applications"
            return None
        else:
            logger.error(f"Unrecognized operating system: {opersys}.")
            return None

    # -- PRIVATE --------------------------------------------------------------
    def freedesktop_create(self) -> tuple[Path | None, Path | None]:
        """Creates shortcut on distributions using FreeDesktop.

        Returns:
            tuple[Path | None, Path | None]: desktop and startmenu path
        """
        # grab shortcut template depending if we are in frozen mode
        # (typically PyInstaller) or as "normal" Python
        if not (getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS")):
            template_shortcut = Path(__file__).parent / "shortcut_freedesktop.template"
            logger.debug(f"Using shortcut template in Python mode: {template_shortcut}")
        else:
            template_shortcut = Path(getattr(sys, "_MEIPASS", sys.executable)).joinpath(
                "shortcuts/shortcut_freedesktop.template"
            )
            logger.debug(
                f"Using shortcut template in packaged mode: {template_shortcut}"
            )

        if not check_path(
            input_path=template_shortcut,
            must_be_a_file=True,
            must_be_readable=True,
            must_exists=True,
            raise_error=False,
        ):
            logger.error(
                FileNotFoundError(
                    f"Shortcut template ({template_shortcut}) doesn't exist. "
                    "Unable to create shortcuts."
                )
            )
            return (None, None)

        # load template
        with template_shortcut.open("r", encoding="UTF-8") as bf_tpl:
            tpl = Template(bf_tpl.read())

        # handle case where work dir is not defined
        if isinstance(self.work_dir, Path):
            profile_name = self.work_dir.name
        else:
            profile_name = "default"

        # subsitute with values
        shortcut_text = tpl.safe_substitute(
            {
                "description": self.description,
                "exec": f"{self.exec_path} {self.exec_arguments}",
                "icon_path": self.icon_path,
                "profile_folder": str(self.work_dir),
                "profile_name": profile_name,
                "shortcut_name": self.name,
            }
        )

        # desktop shortcut
        if self.desktop:
            self.desktop_path.mkdir(parents=True, exist_ok=True)

            # create shortcut
            shortcut_desktop_path = Path(
                self.desktop_path,
                f"qdt.{sluggy(profile_name, '.')}.qgis.desktop",
            )
            shortcut_desktop_path.write_text(shortcut_text, encoding="UTF-8")
            shortcut_desktop_path.chmod(
                shortcut_desktop_path.stat().st_mode
                | (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH),
                follow_symlinks=True,
            )
        else:
            shortcut_desktop_path = None

        # create required shortcut
        if self.start_menu and isinstance(self.startmenu_path, Path):
            self.startmenu_path.mkdir(parents=True, exist_ok=True)
            # create shortcut
            shortcut_start_menu_path = Path(
                self.startmenu_path,
                f"qdt.{sluggy(profile_name, '.')}.qgis.desktop",
            )
            shortcut_start_menu_path.write_text(shortcut_text, encoding="UTF-8")
            shortcut_start_menu_path.chmod(
                shortcut_start_menu_path.stat().st_mode
                | (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH),
                follow_symlinks=True,
            )
        else:
            shortcut_start_menu_path = None

        return (shortcut_desktop_path, shortcut_start_menu_path)

    def win32_create(self) -> tuple[Path | None, Path | None]:
        """Creates shortcut on Windows.

        Returns:
            tuple[Path | None, Path | None]: desktop and startmenu path
        """
        # variable
        _WSHELL = win32com.client.Dispatch("Wscript.Shell", pythoncom.CoInitialize())

        # desktop shortcut
        if self.desktop:
            shortcut_desktop_path = (
                self.desktop_path / f"{self.name}{self.os_config.shortcut_extension}"
            )

            wscript = _WSHELL.CreateShortCut(str(shortcut_desktop_path.resolve()))
            if self.exec_arguments:
                wscript.Arguments = self.exec_arguments
            wscript.Targetpath = str(self.exec_path.resolve())
            if self.work_dir:
                wscript.WorkingDirectory = str(self.work_dir.resolve())
            wscript.WindowStyle = 0
            if self.description:
                wscript.Description = self.description
            else:
                wscript.Description = f"Created by {__title__} {__version__}"
            if self.icon_path:
                if isinstance(self.icon_path, Path):
                    wscript.IconLocation = str(self.icon_path.resolve())
                elif isinstance(self.icon_path, str):
                    wscript.IconLocation = self.icon_path
                else:
                    logger.warning(
                        f"Bad icon path type: {type(self.icon_path)} != (Path, str)."
                    )
            wscript.save()
        else:
            shortcut_desktop_path = None

        # start menu shortcut
        if self.start_menu and isinstance(self.startmenu_path, Path):
            shortcut_start_menu_path = (
                self.startmenu_path / f"{self.name}{self.os_config.shortcut_extension}"
            )

            wscript = _WSHELL.CreateShortCut(str(shortcut_start_menu_path.resolve()))
            if self.exec_arguments:
                wscript.Arguments = self.exec_arguments
            wscript.Targetpath = str(self.exec_path.resolve())
            if self.work_dir is not None:
                wscript.WorkingDirectory = str(self.work_dir.resolve())
            wscript.WindowStyle = 0
            if self.description:
                wscript.Description = self.description
            else:
                wscript.Description = f"Created by {__title__} {__version__}"
            if self.icon_path is not None:
                wscript.IconLocation = str(self.icon_path.resolve())
            wscript.save()
        else:
            shortcut_start_menu_path = None

        return (shortcut_desktop_path, shortcut_start_menu_path)



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_constants
        # for specific test
        python -m unittest tests.test_constants.TestConstants.test_constants
"""


# standard
import tempfile
import unittest
from dataclasses import is_dataclass
from os import environ, getenv, unsetenv
from os.path import expanduser, expandvars
from pathlib import Path
from sys import platform as opersys

# project
from qgis_deployment_toolbelt.constants import (
    OSConfiguration,
    get_qdt_logs_folder,
    get_qdt_working_directory,
)

# ############################################################################
# ########## Classes #############
# ################################


class TestConstants(unittest.TestCase):
    """Test package static variables."""

    def setUp(self) -> None:
        if "QDT_QGIS_EXE_PATH" in environ or getenv("QDT_QGIS_EXE_PATH"):
            unsetenv("QDT_QGIS_EXE_PATH")
            environ.pop("QDT_QGIS_EXE_PATH")

    def test_constants(self):
        """Test types."""

        os_config = OSConfiguration.from_opersys()
        self.assertIsInstance(os_config, OSConfiguration)

        self.assertIsInstance(os_config.qgis_profiles_path, Path)
        self.assertIsInstance(os_config.shortcut_extension, str)
        self.assertIsInstance(os_config.shortcut_forbidden_chars, (tuple, type(None)))
        self.assertIsInstance(os_config.shortcut_icon_extensions, tuple)

        # Check for forbidden characters in the shortcut name
        os_config_forbidden_chars = OSConfiguration(
            names_alter=["test", "fake"],
            name_python=opersys,
            shortcut_forbidden_chars=(" ", "-"),
        )
        self.assertFalse(
            os_config_forbidden_chars.valid_shortcut_name(shortcut_name="qgis-ltr 3.28")
        )
        self.assertTrue(
            os_config_forbidden_chars.valid_shortcut_name(shortcut_name="qgis_ltr_3_28")
        )

    def test_unsupported_operating_system(self):
        """Test that a bad operating system name raise an error."""
        with self.assertRaises(ValueError):
            OSConfiguration.from_opersys("fake_operating_system_name")

    def test_get_qdt_logs_folder(self):
        """Test how QDT logs folder is retrieved"""
        # default value
        self.assertEqual(
            get_qdt_logs_folder(),
            get_qdt_working_directory().joinpath("logs"),
        )

        # using environment variable
        with tempfile.TemporaryDirectory(
            prefix="qdt_test_logs_folder_", ignore_cleanup_errors=True
        ) as tmpdirname:
            environ["QDT_LOGS_DIR"] = tmpdirname
            self.assertEqual(
                get_qdt_logs_folder(),
                Path(tmpdirname),
            )
            unsetenv("QDT_LOGS_DIR")

        # with a bad value set in environment var --> fallback to default value (and error logged)
        environ["QDT_LOGS_DIR"] = f"{Path(__file__).resolve()}"
        self.assertEqual(
            get_qdt_logs_folder(),
            get_qdt_working_directory().joinpath("logs"),
        )
        unsetenv("QDT_LOGS_DIR")

    def test_get_qdt_working_folder(self):
        """Test how QDT working folder is retrieved"""
        # using specific value
        self.assertEqual(
            get_qdt_working_directory(specific_value="~/.cache/qdt/unit-tests/"),
            Path(Path.home(), ".cache/qdt/unit-tests/"),
        )

        # using environment variable
        environ["QDT_LOCAL_WORK_DIR"] = "~/.cache/qdt/unit-tests-env-var/"
        self.assertEqual(
            get_qdt_working_directory(),
            Path(Path.home(), ".cache/qdt/unit-tests-env-var/"),
        )
        unsetenv("QDT_LOCAL_WORK_DIR")

    def test_get_qgis_bin_path(self):
        """Test get GIS exe path helper property"""
        os_config = OSConfiguration.from_opersys()
        # default value
        self.assertEqual(os_config.get_qgis_bin_path, os_config.qgis_bin_exe_path)

    def test_get_qgis_bin_path_with_env_var_str(self):
        """Test with environment var set as str"""
        if "QDT_QGIS_EXE_PATH" in environ:
            environ.pop("QDT_QGIS_EXE_PATH")
        environ["QDT_QGIS_EXE_PATH"] = "/usr/bin/toto"
        os_config = OSConfiguration.from_opersys()
        self.assertEqual(os_config.get_qgis_bin_path, Path("/usr/bin/toto"))

        environ.pop("QDT_QGIS_EXE_PATH")
        environ["QDT_QGIS_EXE_PATH"] = "~/qgis-ltr-bin.exe"
        os_config = OSConfiguration.from_opersys()
        self.assertEqual(
            os_config.get_qgis_bin_path,
            Path(expanduser("~/qgis-ltr-bin.exe")).resolve(),
        )

        environ.pop("QDT_QGIS_EXE_PATH")

    def test_get_qgis_bin_path_with_env_var_dict(self):
        """Test get GIS exe path helper property"""
        if "QDT_QGIS_EXE_PATH" in environ:
            environ.pop("QDT_QGIS_EXE_PATH")
        # with environment var set as dict
        d_test = {
            "linux": "/usr/bin/qgis",
            "darwin": "/usr/bin/qgis",
            "win32": "%PROGRAMFILES%/QGIS/3_22/bin/qgis-ltr-bin.exe",
        }
        environ["QDT_QGIS_EXE_PATH"] = str(d_test)

        os_config = OSConfiguration.from_opersys()

        self.assertEqual(
            os_config.get_qgis_bin_path,
            Path(expandvars(expanduser(d_test.get(opersys)))),
        )

        environ.pop("QDT_QGIS_EXE_PATH")

    def test_get_qgis_profiles_folder_custom(self):
        """Test QGIS profiles path folder."""
        if "QGIS_CUSTOM_CONFIG_PATH" in environ:
            environ.pop("QGIS_CUSTOM_CONFIG_PATH")

        custom_qgis_profiles_folder = Path("tests/fixtures/tmp/custom_qgis_config_path")
        environ["QGIS_CUSTOM_CONFIG_PATH"] = f"{custom_qgis_profiles_folder}"

        os_config: OSConfiguration = OSConfiguration.from_opersys()

        self.assertTrue(is_dataclass(os_config))
        self.assertIsInstance(os_config, OSConfiguration)

        self.assertEqual(
            os_config.qgis_profiles_path,
            custom_qgis_profiles_folder,
        )

        environ.pop("QGIS_CUSTOM_CONFIG_PATH")
        unsetenv("QGIS_CUSTOM_CONFIG_PATH")


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Define toolbelt constant types and values.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# special
from __future__ import annotations

# Standard library
import ast
import logging
from dataclasses import dataclass
from os import PathLike, getenv
from os.path import expanduser, expandvars
from pathlib import Path
from shutil import which
from sys import platform as opersys

# package
from qgis_deployment_toolbelt.utils.check_path import check_path

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# defaults
DEFAULT_QDT_WORKING_FOLDER = Path.home().joinpath(".cache/qgis-deployment-toolbelt")

# QGIS executable filenames
QGIS_BIN_WINDOWS_FILENAME: str = "qgis-bin.exe"
QGIS_LTR_BIN_WINDOWS_FILENAME: str = "qgis-ltr-bin.exe"

# Operating systems
SUPPORTED_OPERATING_SYSTEMS_CODENAMES: tuple[str, ...] = ("darwin", "linux", "win32")


# #############################################################################
# ########## Functions #############
# ##################################


def get_qdt_logs_folder() -> Path:
    """Get QDT logs folder. It uses the default path (a `logs` subfolder under the QDT
        working folder) or the path defined in environment variable `QDT_LOGS_DIR`.

    Returns:
        Path: path to the QDT logs folder.
    """
    # default
    qdt_logs_folder = get_qdt_working_directory().joinpath("logs")

    if isinstance(getenv("QDT_LOGS_DIR"), str):
        qdt_logs_folder_env = Path(expandvars(expanduser(getenv("QDT_LOGS_DIR"))))
        logger.debug(
            f"Logs folder set from QDT_LOGS_DIR environment variable: {qdt_logs_folder_env}"
        )
        # check
        if not check_path(
            input_path=qdt_logs_folder_env,
            must_be_a_file=False,
            must_be_a_folder=True,
            must_be_writable=True,
            raise_error=False,
        ):
            logger.error(
                "Logs folder path set in QDT_LOGS_DIR environment variable is not a "
                f"valid folder path: {qdt_logs_folder_env}. It must to point to a "
                f"writable folder. Fallback to default {qdt_logs_folder}."
            )
        else:
            qdt_logs_folder = qdt_logs_folder_env
    else:
        logger.debug(f"Default value used for QDT logs folder: {qdt_logs_folder}")

    return qdt_logs_folder


def get_qdt_working_directory(
    specific_value: PathLike | None = None, identifier: str | None = None
) -> Path:
    """Get QDT working directory.

    Args:
        specific_value (PathLike, optional): a specific path to use. If set it's \
            expanded and returned. Defaults to None.
        identifier (str, optional): used to make the folder unique. Defaults to None.

    Returns:
        Path: path to the QDT working directory
    """
    if specific_value:
        logger.debug(
            f"QDT working folder - Using the specified value: {specific_value}"
        )
        return Path(expandvars(expanduser(specific_value)))
    elif qdt_local_working_folder := getenv("QDT_LOCAL_WORK_DIR"):
        logger.debug(
            "QDT working folder - Using value specified as environment variable: "
            f"{qdt_local_working_folder}"
        )
        return Path(expandvars(expanduser(qdt_local_working_folder)))
    else:
        if identifier is not None:
            logger.debug(
                f"QDT working folder - Using default path '{DEFAULT_QDT_WORKING_FOLDER}' "
                f"with custom identifier '{identifier}'"
            )
            return Path(
                expandvars(
                    expanduser(
                        getenv(
                            "QDT_LOCAL_WORK_DIR",
                            DEFAULT_QDT_WORKING_FOLDER.joinpath(identifier),
                        ),
                    )
                )
            )
        else:
            logger.debug(
                f"QDT working folder - Using default path: {DEFAULT_QDT_WORKING_FOLDER}"
            )
            return Path(
                expandvars(
                    expanduser(getenv("QDT_LOCAL_WORK_DIR", DEFAULT_QDT_WORKING_FOLDER))
                )
            )


# #############################################################################
# ########## Classes ###############
# ##################################


@dataclass
class OSConfiguration:
    """Settings related to QGIS and depending on operating system"""

    name_python: str
    names_alter: list[str]
    qgis_bin_exe_path: Path | None = None
    qgis_profiles_path: Path | None = None
    shortcut_extension: str | None = None
    shortcut_forbidden_chars: tuple[str, ...] | None = None
    shortcut_icon_extensions: tuple[str, ...] | None = None
    shortcut_icon_default_path: str | None = None

    @property
    def get_qgis_bin_path(self) -> Path:
        """Returns the QGIS path determined from QDT_QGIS_EXE_PATH environment variable,
        or result of which command or fallback to default value passed to the object.

        Returns:
            Path: path to the QGIS bin/exe
        """
        if envvar := getenv("QDT_QGIS_EXE_PATH"):
            if envvar.startswith("{") and envvar.endswith("}"):
                try:
                    qdt_qgis_exe_path = ast.literal_eval(envvar)
                    if isinstance(qdt_qgis_exe_path, dict):
                        logger.debug(
                            f"'QDT_QGIS_EXE_PATH' is a valid dictionary: {envvar}"
                        )
                        for k, v in qdt_qgis_exe_path.items():
                            if k in self.names_alter + [self.name_python]:
                                logger.debug(
                                    f"QGIS path found in 'QDT_QGIS_EXE_PATH' dictionary: {v}"
                                )
                                return Path(expandvars(expanduser(v)))
                except Exception as err:
                    logger.error(
                        f"Failed to interpret 'QDT_QGIS_EXE_PATH' value: {envvar}."
                        f"Trace: {err}"
                    )
            elif check_path(
                input_path=envvar,
                must_exists=False,
                must_be_readable=False,
                raise_error=False,
            ):
                logger.debug(
                    f"'QDT_QGIS_EXE_PATH' is a simple string and a valid path: {envvar}"
                )
                return Path(expandvars(expanduser(envvar)))

            # fallback
            logger.warning(
                f"Unrecognised value format for 'QDT_QGIS_EXE_PATH': {envvar}. "
                "Fallback to default path: "
                f"{Path(expandvars(expanduser(self.qgis_bin_exe_path)))}"
            )
            return Path(expandvars(expanduser(self.qgis_bin_exe_path)))
        elif which_qgis_path := which("qgis"):
            logger.debug(f"QGIS path found using which: {which_qgis_path}")
            return Path(which_qgis_path)
        else:
            logger.debug(
                f"QGIS path not found, using default value: {self.qgis_bin_exe_path}"
            )
            return Path(expandvars(expanduser(self.qgis_bin_exe_path)))

    def valid_shortcut_name(self, shortcut_name: str) -> bool:
        """Check if a given string is a valid shortcut name for the current operating
        system.

        Args:
            shortcut_name (str): given shortcut name to check

        Returns:
            bool: True if the givn string can be used as shortcut name
        """
        if self.shortcut_forbidden_chars is None:
            return True
        for char in self.shortcut_forbidden_chars:
            if char in shortcut_name:
                logger.error(
                    f"Shortcut name '{shortcut_name}' contains forbidden char '{char}'"
                )
                return False
        return True

    @classmethod
    def from_opersys(
        cls,
        operating_system_codename: str | None = None,
    ) -> OSConfiguration:
        """Create configuration object with defaults values from a operating system
            code name.

        Args:
            operating_system_codename: operating system code name as specified in \
                sys.platform. If None, fallback to current operating system. \
                    Defaults to None.

        Returns:
            Self: OSConfiguration object with defaults settings
        """
        # if not specified, fallback to current operating system
        if operating_system_codename is None:
            operating_system_codename = opersys
            logger.debug(
                f"Getting configuration for current operating system: {opersys}"
            )

        # returning configuration for operating system
        if operating_system_codename == "darwin":
            return cls(
                name_python="darwin",
                names_alter=["apple", "mac", "macos"],
                qgis_bin_exe_path=Path("/usr/bin/qgis"),
                qgis_profiles_path=Path(
                    getenv(
                        "QGIS_CUSTOM_CONFIG_PATH",
                        Path.home()
                        / "Library/Application Support/QGIS/QGIS3/profiles/",
                    )
                ),
                shortcut_extension="app",
                shortcut_icon_extensions=("icns",),
            )
        elif operating_system_codename == "linux":
            return cls(
                name_python="linux",
                names_alter=["kubuntu", "ubuntu"],
                qgis_bin_exe_path=Path("/usr/bin/qgis"),
                qgis_profiles_path=Path(
                    getenv(
                        "QGIS_CUSTOM_CONFIG_PATH",
                        Path.home() / ".local/share/QGIS/QGIS3/profiles/",
                    )
                ),
                shortcut_extension=".desktop",
                shortcut_icon_extensions=("ico", "svg", "png"),
                shortcut_icon_default_path="qgis",
            )
        elif operating_system_codename == "win32":
            return cls(
                name_python="win32",
                names_alter=["win", "windows"],
                qgis_bin_exe_path=Path(
                    expandvars(
                        expanduser("%PROGRAMFILES%/QGIS 3.34.4/bin/qgis-ltr-bin.exe")
                    )
                ),
                qgis_profiles_path=Path(
                    getenv(
                        "QGIS_CUSTOM_CONFIG_PATH",
                        expandvars("%APPDATA%/QGIS/QGIS3/profiles"),
                    )
                ),
                shortcut_extension=".lnk",
                shortcut_forbidden_chars=("<", ">", ":", '"', "/", "\\", "|", "?", "*"),
                shortcut_icon_extensions=("ico",),
            )
        else:
            raise ValueError(
                f"Unsupported operating system specified: {operating_system_codename}. "
                "Must be one of: {', '.join('darwin', 'linux', 'win32')}"
            )



#! python3  # noqa: E265

"""
    Manage application shortcuts on end-user machine.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from pathlib import Path

# package
from qgis_deployment_toolbelt.__about__ import __title__, __version__
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.profiles.qdt_profile import QdtProfile
from qgis_deployment_toolbelt.shortcuts.shortcuts_handler import ApplicationShortcut

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobShortcutsManager(GenericJob):
    """
    Job to create or remove shortcuts on end-user machine.
    """

    ID: str = "shortcuts-manager"
    OPTIONS_SCHEMA: dict = {
        "action": {
            "type": str,
            "required": False,
            "default": "create_or_restore",
            "possible_values": ("create", "create_or_restore", "remove"),
            "condition": "in",
        },
        "include": {
            "type": (list, str),
            "required": False,
            "default": None,
            "possible_values": None,
            "condition": None,
            "sub_options": {
                "additional_arguments": {
                    "type": str,
                    "required": False,
                    "default": None,
                    "possible_values": None,
                    "condition": None,
                },
                "desktop": {
                    "type": bool,
                    "required": False,
                    "default": False,
                    "possible_values": None,
                    "condition": None,
                },
                "icon": {
                    "type": str,
                    "required": False,
                    "default": None,
                    "possible_values": None,
                    "condition": None,
                },
                "label": {
                    "type": str,
                    "required": False,
                    "default": "QGIS",
                    "possible_values": None,
                    "condition": None,
                },
                "profile": {
                    "type": str,
                    "required": True,
                    "default": None,
                    "possible_values": None,
                    "condition": None,
                },
                "start_menu": {
                    "type": bool,
                    "required": False,
                    "default": True,
                    "possible_values": None,
                    "condition": None,
                },
            },
        },
    }
    SHORTCUTS_CREATED: list = []
    SHORTCUTS_REMOVED: list = []

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        :param dict options: profiles source (remote, can be a local network) and
        destination (local).
        """
        super().__init__()
        self.options: dict = self.validate_options(options)

    def run(self) -> None:
        """Execute job logic."""
        # check of there are some profiles folders within the downloaded folder
        downloaded_profiles = self.list_downloaded_profiles()
        if downloaded_profiles is None:
            logger.error("No QGIS profile found in the downloaded folder.")
            return

        # check action
        if self.options.get("action") in ("create", "create_or_restore"):
            for inc_shortcut in self.options.get("include", []):
                qdt_profile = self.get_matching_profile_from_name(
                    li_profiles=downloaded_profiles,
                    profile_name=inc_shortcut.get("profile"),
                )
                if not qdt_profile:
                    continue

                # instanciate shortcut
                shortcut = ApplicationShortcut(
                    name=inc_shortcut.get("label"),
                    exec_path=self.os_config.get_qgis_bin_path,
                    description=f"Created with {__title__} {__version__}",
                    icon_path=self.get_icon_path(
                        profile=qdt_profile, icon_filename=inc_shortcut.get("profile")
                    ),
                    exec_arguments=self.get_arguments_ready(
                        inc_shortcut.get("profile"),
                        inc_shortcut.get("additional_arguments"),
                    ),
                    work_dir=qdt_profile.path_in_qgis,
                )
                shortcut.create(
                    desktop=inc_shortcut.get("desktop", False),
                    start_menu=inc_shortcut.get("start_menu", True),
                )
                logger.info(f"Created shortcut {shortcut.name}")
                self.SHORTCUTS_CREATED.append(shortcut)

            logger.debug(f"Job {self.ID} ran successfully.")
        else:
            raise NotImplementedError

    # -- INTERNAL LOGIC ------------------------------------------------------
    def get_matching_profile_from_name(
        self, li_profiles: list[QdtProfile], profile_name: str
    ) -> QdtProfile:
        """Get a profile from list of profiles using a profile's name to match.

        Args:
            li_profiles (list[QdtProfile]): list of profile to look into
            profile_name (str): profile name

        Returns:
            QdtProfile: matching profile object
        """
        # load profile
        matching_qdt_profile = [
            pr for pr in li_profiles if profile_name in (pr.name, pr.folder.name)
        ]
        if not len(matching_qdt_profile):
            logger.error(
                "Unable to get a matching profile among downloaded ones with "
                f"the name: {profile_name}"
            )
            return None

        qdt_profile = matching_qdt_profile[0]
        logger.info(
            f"Downloaded profile matched: {qdt_profile.name} from "
            f"{qdt_profile.folder}"
        )
        return qdt_profile

    def get_icon_path(self, profile: QdtProfile, icon_filename: str) -> Path | None:
        """Get icon path to use with the shortcut. Looking for:

            1. checks if an icon key has been specified in the profile.json file and
                if it exists
            2. checks if an icon has been specified into scenario and if it exists
                within profile folder (in QGIS)
            3. checks if an icon has been specified into scenario and if it exists under
                a subfolder starting from the toolbelt (and handling pathlib OSError)

        Args:
            profile (QdtProfile): QGIS profile object
            icon_filename (str): icon path as mentioned into the scenario file

        Returns:
            Path: icon path as Path if str or Path, else None
        """
        profile_icon_installed = None

        # 1. check icon specified in profile.json
        if profile.icon:
            profile_icon_installed = profile.path_in_qgis.joinpath(profile.icon)
            if profile_icon_installed.is_file():
                logger.debug(
                    f"Icon set in profile.json exists so it will be used: "
                    f"{profile_icon_installed}"
                )
                return profile_icon_installed.resolve()

        # if not specified in profile.json nor in scenario --> None
        if icon_filename is None:
            logger.debug("No icon found in profile.json nor in scenario.")
            return None

        # 2. check if icon specified in scenario exists in profile folder
        try:
            li_icons_sub_profile_folder = list(
                profile.path_in_qgis.rglob(f"{icon_filename}")
            )
            if len(li_icons_sub_profile_folder):
                logger.debug(
                    "Icon found under the installed profile folder: "
                    f"{li_icons_sub_profile_folder[0].resolve()}"
                )
                return li_icons_sub_profile_folder[0].resolve()
        except OSError as err:
            logger.error(f"Looking for icon into profile subfolder failed. {err}")

        # 3. check if icon specified in scenario exists under QDT
        try:
            li_icons_sub_qdt_folder = list(Path(".").glob(f"{icon_filename}"))
            if len(li_icons_sub_qdt_folder):
                logger.debug(
                    "Icon found under the QDT folder: "
                    f"{li_icons_sub_qdt_folder[0].resolve()}"
                )
                return li_icons_sub_qdt_folder[0].resolve()
        except OSError as err:
            logger.error(f"Looking for icon into profile subfolder failed. {err}")

        return profile_icon_installed

    def get_arguments_ready(
        self, profile: str, in_arguments: str | None = None
    ) -> list[str]:
        """Prepare arguments for the executable shortcut.

        Args:
            profile (str): profile's name
            in_arguments (str, optional): argument as defined in the scenario file. Defaults to None.

        Returns:
            list[str]: tuple of strings separated by spaces
        """
        # add profile name
        arguments: list = ["--profile", f'"{profile}"']

        # add additional arguments
        if in_arguments:
            arguments.extend(in_arguments.split(" "))

        return arguments


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Small module to get network proxies configuration.

    Author: Julien Moura (github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from functools import lru_cache
from os import environ
from urllib.request import getproxies

# package
from qgis_deployment_toolbelt.utils.url_helpers import check_str_is_url

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################
@lru_cache
def get_proxy_settings() -> dict:
    """Retrieves network proxy settings from operating system configuration or
    environment variables.

    Returns:
        dict: proxy settings with protocl as key and URL as value
    """
    proxy_settings = {}
    if environ.get("QDT_PROXY_HTTP"):
        proxy_settings = {
            "http": environ.get("QDT_PROXY_HTTP"),
            "https": environ.get("QDT_PROXY_HTTP"),
        }
        logger.info(
            "Proxies settings from custom QDT in environment vars (QDT_PROXY_HTTP): "
            f"{proxy_settings}"
        )
    elif environ.get("HTTP_PROXY") or environ.get("HTTPS_PROXY"):
        if environ.get("HTTP_PROXY") and environ.get("HTTPS_PROXY"):
            proxy_settings = {
                "http": environ.get("HTTP_PROXY"),
                "https": environ.get("HTTPS_PROXY"),
            }
            logger.info(
                "Proxies settings from generic environment vars (HTTP_PROXY "
                f"and HTTPS_PROXY): {proxy_settings}"
            )
        elif environ.get("HTTP_PROXY") and not environ.get("HTTPS_PROXY"):
            proxy_settings = {
                "http": environ.get("HTTP_PROXY"),
            }
            logger.info(
                "Proxies settings from generic environment vars (HTTP_PROXY only): "
                f"{proxy_settings}"
            )
        elif not environ.get("HTTP_PROXY") and environ.get("HTTPS_PROXY"):
            proxy_settings = {
                "https": environ.get("HTTPS_PROXY"),
            }
            logger.info(
                "Proxies settings from generic environment vars (HTTPS_PROXY only): "
                f"{proxy_settings}"
            )
    elif getproxies():
        proxy_settings = getproxies()
        logger.debug(f"Proxies settings found in the OS: {proxy_settings}")
    else:
        logger.debug("No proxy settings found in environment vars nor OS settings.")

    # check scheme and URL validity
    if isinstance(proxy_settings, dict):
        for scheme, proxy_url in proxy_settings.items():
            if not check_str_is_url(input_str=proxy_url, raise_error=False):
                logger.warning(
                    f"Proxy value for {scheme} is not a valid URL: {proxy_url}. Can "
                    "lead to troubles."
                )

    return proxy_settings


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa E265

"""Usage from the repo root folder:

    .. code-block:: python

        # for whole test
        python -m unittest tests.test_scenario_reader
        # for specific
        python -m unittest tests.test_scenario_reader.TestScenarioReader.test_load_from_yaml
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import unittest
from pathlib import Path

# module target
from qgis_deployment_toolbelt.scenarios.scenario_reader import ScenarioReader

# #############################################################################
# ########## Classes ###############
# ##################################


class TestScenarioReader(unittest.TestCase):
    """Test module."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""
        cls.good_scenario_files = sorted(
            Path("tests/fixtures/").glob("scenarios/good_*.y*ml")
        )

    # standard methods
    def setUp(self):
        """Fixtures prepared before each test."""
        pass

    def tearDown(self):
        """Executed after each test."""
        pass

    # -- TESTS ---------------------------------------------------------
    def test_load_from_yaml(self):
        """Test YAML loader"""
        for i in self.good_scenario_files:
            reader = ScenarioReader(in_yaml=i)
            self.assertIsInstance(reader.scenario, dict)

            # scenario sections
            self.assertIn("settings", reader.scenario)
            self.assertIn("metadata", reader.scenario)
            self.assertIn("steps", reader.scenario)

            # validation
            validation = reader.validate_scenario()
            self.assertIsInstance(validation, tuple)
            self.assertEqual(len(validation), 2)
            self.assertIsInstance(validation[0], bool)
            self.assertIsInstance(validation[1], list)

            # properties
            self.assertIsInstance(reader.metadata, dict)
            self.assertIsInstance(reader.settings, dict)
            self.assertIsInstance(reader.steps, list)

    def test_missing_scenario(self):
        """Test missing scenario."""
        with self.assertRaises(TypeError):
            ScenarioReader()

        with self.assertRaises(TypeError):
            yaml_as_dict = {"metadata": {"name": "test"}}
            ScenarioReader(yaml_as_dict)



#! python3  # noqa: E265

"""Helpers to format text and variables."""


# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import logging
from functools import lru_cache
from math import floor
from math import log as math_log

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################


@lru_cache
def convert_octets(octets: int) -> str:
    """Convert a mount of octets in readable size.

    Args:
        octets: mount of octets to convert

    Returns:
        size in a human readable format: ko, Mo, etc.

    Example:

    .. code-block:: python

        >>> convert_octets(1024)
        1 ko
        >>> from pathlib import Path
        >>> convert_octets(Path(my_file.txt).stat().st_size)
    """
    # check zero
    if octets == 0:
        return "0 octet"

    # conversion
    size_name = ("octets", "Ko", "Mo", "Go", "To", "Po")
    i = int(floor(math_log(octets, 1024)))
    p = pow(1024, i)
    s = round(octets / p, 2)

    return f"{s} {size_name[i]}"


@lru_cache
def url_ensure_trailing_slash(in_url: str) -> str:
    """Make sure an URL has a trailing slash.

    Args:
        in_url (str): input URL

    Returns:
        str: URL with trailing slash added (only if not already present)
    """
    if in_url.endswith("/"):
        logger.debug(f"URL already had a trailing slash: {in_url}")
        return in_url
    else:
        logger.debug(f"A trailing slash has been added to input URL: {in_url}/")
        return f"{in_url}/"



#! python3  # noqa: E265

"""
    Helpers to check file: readable, exists, etc..

    Author: Julien Moura (https://github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from urllib.parse import urlparse

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Functions #############
# ##################################


def check_str_is_url(
    input_str: str,
    ref_shemes: tuple = ("http", "https"),
    raise_error: bool = True,
):
    """Checks if a given str is a valid URL.

    Args:
        input_str (str): The input str to check.
        ref_schemes (tuple, optional): The reference schemes for valid URLs.
            By default, it includes ("http", "https").
        raise_error (bool, optional): Indicates whether an exception should be raised
            for invalid str. Default is True.

    Returns:
        bool: True if the str is a valid URL, False otherwise.

    Raises:
        ValueError: If the str is not a valid URL and `raise_error` is True.
        TypeError: If an error occurs during str checking and `raise_error` is True.
    """
    # convert into str
    if not isinstance(input_str, str):
        logger.warning(
            f"{input_str} is not a str but {type(input_str)}. Take care, "
            "the conversion is not safe..."
        )
        input_str = str(input_str)

    try:
        parsed_url = urlparse(input_str)
        if parsed_url.scheme in ref_shemes and parsed_url.netloc:
            logger.debug(f"{input_str} is a valid URL.")
            return True
        else:
            error_message = f"{input_str} is not a valid URL."
            if raise_error:
                raise ValueError(error_message)
            logger.debug(error_message)
            return False
    except ValueError as err:
        error_message = (
            f"{input_str} is not a valid URL. An error occurred during "
            f"check. Trace : {err}"
        )
        if raise_error:
            raise TypeError(error_message)
        logger.warning(error_message)
        return False


# ############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Utilities to manage user security groups either on Linux and Windows.

    Author: Julien Moura (https://github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# standard library
import logging
import subprocess
from functools import lru_cache
from getpass import getuser
from platform import uname
from sys import platform as opersys

# Imports depending on operating system
if opersys == "win32":
    """windows"""

    # 3rd party
    import pyad
    import win32com
    import win32net
    from pywintypes import error as PyWinException

else:
    import grp

# package
from qgis_deployment_toolbelt.utils.win32utils import (
    ExtendedNameFormat,
    get_current_user_extended_data,
)

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################


@lru_cache
def get_user_active_directory_object() -> "pyad.ADUser | None":
    """Get user as an Active Directory object.

    Returns:
        pyad.ADUser | None: user object or None if something went wrong
    """
    user_obj = None
    user_guid_or_name = get_user_name_or_guid()
    try:
        if user_guid_or_name.startswith("{"):
            user_obj = pyad.aduser.ADUser.from_guid(user_guid_or_name)
        else:
            user_obj = pyad.aduser.ADUser.from_cn(user_guid_or_name)
    except Exception as err:
        logger.error(f"Unable to get the user object. Trace: {err}")
    return user_obj


@lru_cache
def get_user_name_or_guid() -> str:
    """Get user GUID or name as fallback.

    Returns:
        str: user GUID or name (getpass.getuser())
    """
    if opersys.lower() in ("win32", "windows"):
        try:
            user_guid_or_name = get_current_user_extended_data(
                ExtendedNameFormat.NameUniqueId
            )
            logger.debug(
                f"Using user GUID to retrieve domain groups: {user_guid_or_name}"
            )
        except Exception as err:
            logger.info(
                f"Unable to retrieve user GUID. Fallback to user name. Trace: {err}"
            )
    else:
        user_guid_or_name = getuser()
        logger.debug(f"Using username to retrieve domain groups: {user_guid_or_name}")

    return user_guid_or_name


@lru_cache
def get_user_local_groups(user_name: str | None = None) -> list[str]:
    """Lists the local groups to which the user belong.

    On Linux and MacOS, it relies on the grp module (only available).
    On Windows, it relies on win32net (COM).

    Args:
        user_name (str | None, optional): name of user. If None, the current user name
            is used. Defaults to None.

    Raises:
        NotImplementedError: if operating system is not supported.

    Returns:
        list[str]: sorted list of unique groups names
    """
    if user_name is None:
        user_name = getuser()

    if opersys.lower() in ("darwin", "linux"):
        return sorted({g.gr_name for g in grp.getgrall() if user_name in g.gr_mem})
    elif opersys.lower() in ("win32", "windows"):
        server_host_name = uname()[1]
        try:
            local_groups = sorted(
                set(win32net.NetUserGetLocalGroups(server_host_name, user_name))
            )
        except PyWinException as err:
            logger.info(
                f"Retrieving user ('{user_name}') local groups on {server_host_name} "
                "failed. This usually means that it's not a local session, and that the "
                "computer is linked to a domain and subscribed to a directory. "
                f"Trace: {err}"
            )
            local_groups = []
        return local_groups
    else:
        raise NotImplementedError(f"Unsupported operating system: {opersys}")


@lru_cache
def get_user_domain_groups(user_guid_or_name: str | None = None) -> list[str]:
    """Lists the domain groups to which the user belong.

    On Linux and MacOS, it always return an empty list. TODO: implement it (probably
        using pure LDAP).
    On Windows, it relies on win32net (COM).

    Args:
        user_guid_or_name (str | None, optional): name of user. If None, the current user name
            is used. Defaults to None.

    Raises:
        NotImplementedError: if operating system is not supported.

    Returns:
        list[str]: sorted list of unique groups names
    """
    if user_guid_or_name is None:
        user_guid_or_name = get_user_name_or_guid()

    if not is_computer_attached_to_a_domain():
        logger.debug(
            "Computer is not attached to a domain so retrieving user's domain groups "
            "is meaningless. Returning empty list."
        )
        return []

    if opersys.lower() in ("darwin", "linux"):
        # TODO: find a way to retrive user's domain groups on Linux and MacOS (probably
        #  using pure ldap)
        return []
    elif opersys.lower() in ("win32", "windows"):
        user_obj = get_user_active_directory_object()
        user_groups: list[pyad.ADGroup] | None = user_obj.get_memberOfs()
        if not user_groups or not len(user_groups):
            logger.warning(
                f"It looks like '{user_guid_or_name}' does not belong to any domain group."
            )
            return []
        else:
            logger.info(
                f"The current user '{user_guid_or_name}' belongs to {len(user_groups)} Active "
                "Directory groups."
            )

        return sorted(
            {
                grp.get_attribute("name")[0]
                for grp in user_groups
                if isinstance(grp, pyad.ADGroup)
            }
        )
    else:
        raise NotImplementedError(f"Unsupported operating system: {opersys}")


def _is_computer_in_domain_powershell() -> bool:
    """Check if the computer is joined to a domain or a workgroup using PowerShell.

    Returns:
        bool: True if the computer is joined to a domain or workgroup, False otherwise.
    """
    # use PowerShell to retrieve domain information
    domain = subprocess.run(
        ["powershell.exe", "(Get-CimInstance Win32_ComputerSystem).Domain"],
        stdout=subprocess.PIPE,
        text=True,
    ).stdout.strip()

    # check if domain is different from workgroup
    return domain.lower() != "workgroup"


def _is_computer_in_domain_pyad() -> bool:
    """Check if the computer is joined to a domain or a workgroup using PyAD.

    Returns:
        bool: True if the computer is joined to a domain or workgroup, False otherwise.
    """
    try:
        user_obj = get_user_active_directory_object()
        return bool(len(user_obj.get_attribute("memberOf")))
    except Exception as err:
        logger.info(
            "Based on pyad, the current computer is not attached to any domain."
            f" Trace: {err}"
        )
        return False


def _is_computer_in_domain_win32() -> bool:
    """Check if the computer is joined to a domain or a workgroup using win32com.

    Returns:
        bool: True if the computer is joined to a domain or workgroup, False otherwise.
    """
    # try to import wmi
    try:
        wmi = win32com.client.GetObject("winmgmts://./root/cimv2")
    except Exception as err:
        logging.info(
            "Unable to load WMI (Windows Management Instrumentation) from "
            f"win32com.client. Other options will be used. Trace: {err}"
        )
        raise err

    try:
        # Execute a query to retrieve information about the domain
        query = "SELECT * FROM Win32_ComputerSystem"
        result = wmi.ExecQuery(query)

        # Check if the domain attribute is not None
        for item in result:
            if item.Domain and item.Domain.lower() != "workgroup":
                logger.debug(f"Computer is joined to a domain: {item.Domain.lower()}")
                return True

    except Exception as err:
        logger.error(f"Getting domain from WMI through win32com failed. Trace: {err}")
        raise err

    return False


@lru_cache
def is_computer_attached_to_a_domain() -> bool:
    """Determine if the computer is attached to a domain or not.

    On Linux and MacOS, it always return False.
    On Windows, it tries to use wmi (Windows Management Instrumentation), Active
        Directory (through pyad) or subprocessed powershell as final fallback.

    Raises:
        NotImplementedError: if operating system is not supported.

    Returns:
        bool: True if the computer is attached to a domain.
    """
    if opersys.lower() in ("darwin", "linux"):
        # TODO: find a way to determine if a computer is attached to a domain on Linux and MacOS
        return False
    elif opersys.lower() in ("win32", "windows"):
        try:
            logger.debug(
                "Determine if computer joined to a domain using WMI through win32 API."
            )
            return _is_computer_in_domain_win32()
        except Exception as err:
            logger.error(
                f"Something went wrong using WMI through win32com. Trace: {err}. "
                "Trying with other methods..."
            )

        try:
            logger.debug("Determine if computer joined to a domain using PyAD.")
            return _is_computer_in_domain_pyad()
        except Exception as err:
            logger.error(
                f"Something went wrong using PyAD. Trace: {err}. "
                "Fallback to PowerShell..."
            )
        # fallback with PowerShell
        return _is_computer_in_domain_powershell()

    else:
        raise NotImplementedError(f"Unsupported operating system: {opersys}")


@lru_cache
def is_user_in_group(group_name: str, user_name: str | None = None) -> bool:
    """Check if a user is a member of a specified group.

    Args:
        user_name (str | None, optional): The username to check. If None, the current
            user name is used. Defaults to None.
        group_name (str): The name of the group to check membership.

    Returns:
        bool: True if the user is a member of the group, False otherwise or if
            something fails.
    """
    # local variables
    local_groups = []
    domain_groups = []

    # if user not specified, use the current user name
    if user_name is None:
        user_name = getuser()

    # first, get local groups
    try:
        local_groups = get_user_local_groups()
        logger.debug(
            f"User '{user_name}' belongs to following LOCAL groups: "
            f"{'; '.join(local_groups)}. "
        )
    except Exception as err:
        logger.error(f"Retrieving user's LOCAL groups failed. Trace: {err}")

    # if computer is not attached to a domain, check only on local groups
    if not is_computer_attached_to_a_domain():
        return group_name in local_groups

    try:
        domain_groups = get_user_domain_groups()
        logger.debug(
            f"User '{user_name}' belongs to following DOMAIN groups: "
            f"{'; '.join(domain_groups)}. "
        )
    except Exception as err:
        logger.error(f"Retrieving user's DOMAIN groups failed. Trace: {err}")

    return any([group_name in domain_groups, group_name in local_groups])



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_str2bool
        # for specific test
        python -m unittest tests.test_utils_str2bool.TestUtilsStrToBool.test_str2bool
"""

# standard library
import unittest

# project
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsStrToBool(unittest.TestCase):
    """Test package utilities."""

    def test_str2bool(self):
        """Test str2bool."""
        # OK
        self.assertTrue(str2bool("True"))
        self.assertTrue(str2bool(True))
        self.assertFalse(str2bool("False", raise_exc=True))
        self.assertIsNone(str2bool("faux", raise_exc=False))

        # KO
        with self.assertRaises(TypeError):
            str2bool(input_var=int(10), raise_exc=True)
        self.assertIsNone(str2bool(input_var=int(10), raise_exc=False))

        with self.assertRaises(ValueError):
            str2bool(input_var="vrai", raise_exc=True)


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Download plugins listed into profiles.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from concurrent.futures import ThreadPoolExecutor
from os import getenv
from pathlib import Path
from shutil import copy2

# package
from qgis_deployment_toolbelt.__about__ import __title_clean__
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.plugins.plugin import QgisPlugin
from qgis_deployment_toolbelt.utils.check_path import check_path
from qgis_deployment_toolbelt.utils.file_downloader import download_remote_file_to_local
from qgis_deployment_toolbelt.utils.str2bool import str2bool

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobPluginsDownloader(GenericJob):
    """
    Job to download plugins.
    """

    ID: str = "qplugins-downloader"
    OPTIONS_SCHEMA: dict = {
        "force": {
            "type": bool,
            "required": False,
            "default": False,
            "possible_values": None,
            "condition": None,
        },
        "threads": {
            "type": int,
            "required": False,
            "default": 5,
            "possible_values": (1, 2, 3, 4, 5),
            "condition": "in",
        },
    }

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        :param dict options:  job options.
        """
        super().__init__()
        self.options: dict = self.validate_options(options)

        # where QDT downloads plugins
        self.qdt_plugins_folder.mkdir(exist_ok=True, parents=True)
        logger.info(f"QDT plugins folder: {self.qdt_plugins_folder}")

    def run(self) -> None:
        """Execute job logic."""
        # list plugins through different profiles
        qdt_referenced_plugins = self.list_referenced_plugins(
            parent_folder=self.qdt_downloaded_repositories
        )
        if qdt_referenced_plugins is None or not len(qdt_referenced_plugins):
            logger.info(
                f"No plugin found in profile.json files within {self.qdt_working_folder}"
            )
            return

        # filter plugins to download, filtering out those which are not already present locally
        if self.options.get("force") is True:
            qdt_plugins_to_download = qdt_referenced_plugins
            qdt_plugins_to_copy = []
        else:
            qdt_plugins_to_download = self.filter_list_downloadable_plugins(
                input_list=qdt_referenced_plugins
            )
            qdt_plugins_to_copy = self.filter_list_copiable_plugins(
                input_list=qdt_referenced_plugins
            )
            if not len(qdt_plugins_to_download + qdt_plugins_to_copy):
                logger.info(
                    f"All referenced plugins are already present in {self.qdt_plugins_folder}. "
                    "Skipping download/copy step."
                )
                return

        # launch download
        if len(qdt_plugins_to_download):
            downloaded_plugins, failed_downloads = self.download_remote_plugins(
                plugins_to_download=qdt_plugins_to_download,
                destination_parent_folder=self.qdt_plugins_folder,
                threads=self.options.get("threads", 5),
            )
            logger.debug(f"{len(downloaded_plugins)} plugins downloaded.")
            if len(failed_downloads):
                logger.error(
                    f"{len(failed_downloads)} failed plugin downloads. "
                    "Check previous log lines."
                )

        if len(qdt_plugins_to_copy):
            copied_plugins, failed_copies = self.copy_plugins(
                plugins_to_copy=qdt_plugins_to_copy,
                destination_parent_folder=self.qdt_plugins_folder,
            )
            logger.debug(f"{len(copied_plugins)} plugins copied.")
            if len(failed_copies):
                logger.error(
                    f"{len(failed_copies)} failed plugin copies. "
                    "Check previous log lines."
                )

        logger.debug(f"Job {self.ID} ran successfully.")

    def copy_plugins(
        self, plugins_to_copy: list[QgisPlugin], destination_parent_folder: Path
    ) -> tuple[list[QgisPlugin], list[QgisPlugin]]:
        """Copy listed plugins into the specified folder.

        Args:
            plugins_to_copy (List[QgisPlugin]): list of plugins to copy
            destination_parent_folder (Path): where to store copied plugins

        Returns:
            Tuple[List[QgisPlugin],List[QgisPlugin]]: tuple of (copied plugins, failed copies)
        """
        copied_plugins: list[QgisPlugin] = []
        failed_plugins: list[QgisPlugin] = []

        logger.debug(
            f"Copying {len(plugins_to_copy)} plugins into {destination_parent_folder}."
        )
        for plugin in plugins_to_copy:
            # check if source plugin can be accessed
            try:
                check_path(
                    input_path=plugin.url,
                    must_be_a_file=True,
                    must_be_readable=True,
                    must_be_a_folder=False,
                    must_exists=True,
                )
                src_plugin_path = Path(plugin.url)
            except Exception as err:
                logger.error(
                    f"The plugin '{plugin.name}' can't be copied from {plugin.url}. "
                    f"Trace: {err}."
                )
                failed_plugins.append(plugin)
                continue

            # try to copy
            try:
                dst_plugin_path = destination_parent_folder / src_plugin_path.name
                copy2(src=src_plugin_path, dst=destination_parent_folder)
                logger.info(
                    f"Plugin {plugin.name} has been copied from {src_plugin_path} "
                    f"to {dst_plugin_path}"
                )
                check_path(
                    input_path=dst_plugin_path,
                    must_be_a_file=True,
                    must_exists=True,
                    raise_error=True,
                )

                dst_plugin_path_final = dst_plugin_path.rename(
                    dst_plugin_path.with_name(f"{plugin.id_with_version}.zip")
                )
                logger.debug(
                    f"Plugin ZIP archive has been renamed from '{dst_plugin_path}' "
                    f"into '{dst_plugin_path_final}' to make it consistent with sync."
                )

                copied_plugins.append(plugin)
            except Exception as err:
                logger.error(
                    f"Copy of plugin {plugin.name} from {plugin.url} failed. Trace: {err}"
                )
                failed_plugins.append(plugin)
                continue

        return copied_plugins, failed_plugins

    def download_remote_plugins(
        self,
        plugins_to_download: list[QgisPlugin],
        destination_parent_folder: Path,
        threads: int = 5,
    ) -> tuple[list[QgisPlugin], list[QgisPlugin]]:
        """Download listed plugins into the specified folder, using multithreads or not.

        Args:
            plugins_to_download (List[QgisPlugin]): list of plugins to download
            destination_parent_folder (Path): where to store downloaded plugins
            threads (int, optional): number of threads to use. If 0, downloads will be \
                performed synchronously. Defaults to 5.

        Returns:
            Tuple[List[QgisPlugin],List[QgisPlugin]]: tuple of \
                (downloaded plugins, failed downloads)
        """
        downloaded_plugins: list[QgisPlugin] = []
        failed_plugins: list[QgisPlugin] = []

        if threads < 2:
            logger.debug(
                f"Downloading {len(plugins_to_download)} plugins in a single thread."
            )
            for plugin in plugins_to_download:
                # local path
                plugin_download_path = Path(
                    destination_parent_folder, f"{plugin.id_with_version}.zip"
                )
                try:
                    download_remote_file_to_local(
                        local_file_path=plugin_download_path,
                        remote_url_to_download=plugin.download_url,
                        content_type="application/zip",
                        use_stream=str2bool(getenv("QDT_STREAMED_DOWNLOADS", True)),
                    )
                    logger.info(
                        f"Plugin {plugin.name} from {plugin.download_url} "
                        f"downloaded in {plugin_download_path}"
                    )
                    downloaded_plugins.append(plugin)
                except Exception as err:
                    logger.error(
                        f"Download of plugin {plugin.name} failed. Trace: {err}"
                    )
                    failed_plugins.append(plugin)
                    continue
        else:
            logger.debug(
                f"Downloading {len(plugins_to_download)} plugins in {threads} threads."
            )
            with ThreadPoolExecutor(
                max_workers=threads, thread_name_prefix=f"{__title_clean__}"
            ) as executor:
                for plugin in plugins_to_download:
                    # local path
                    plugin_download_path = Path(
                        destination_parent_folder, f"{plugin.id_with_version}.zip"
                    )

                    # submit download to pool
                    try:
                        executor.submit(
                            # func to execute
                            download_remote_file_to_local,
                            # func parameters
                            local_file_path=plugin_download_path,
                            remote_url_to_download=plugin.download_url,
                            content_type="application/zip",
                            use_stream=str2bool(getenv("QDT_STREAMED_DOWNLOADS", True)),
                        )
                        downloaded_plugins.append(plugin)
                    except Exception as err:
                        logger.error(err)
                        failed_plugins.append(plugin)

        return downloaded_plugins, failed_plugins

    def list_referenced_plugins(self, parent_folder: Path) -> list[QgisPlugin] | None:
        """Return a list of plugins referenced in profile.json files found within a \
            parent folder and sorted by unique id with version.

        Args:
            parent_folder (Path): folder to start searching for profile.json files

        Returns:
            List[QgisPlugin]: list of plugins referenced within profile.json files
        """
        unique_plugins_identifiers: list = []
        all_profiles: list[QgisPlugin] = []

        # check of there are some profiles folders within the downloaded folder
        li_qdt_downloaded_profiles = self.list_downloaded_profiles()
        if li_qdt_downloaded_profiles is None:
            logger.error("No QGIS profile found in the downloaded folder.")
            return

        for qdt_profile in li_qdt_downloaded_profiles:
            for plugin in qdt_profile.plugins:
                if plugin.id_with_version not in unique_plugins_identifiers:
                    unique_plugins_identifiers.append(plugin.id_with_version)
                    all_profiles.append(plugin)

        logger.debug(
            f"{len(unique_plugins_identifiers)} unique plugins referenced in "
            f"{len(li_qdt_downloaded_profiles)} profiles in {parent_folder.resolve()}: "
            f"{','.join(sorted(unique_plugins_identifiers))}"
        )
        return sorted(all_profiles, key=lambda x: x.id_with_version)

    def filter_list_downloadable_plugins(
        self, input_list: list[QgisPlugin]
    ) -> list[QgisPlugin]:
        """Filter input list of plugins keeping only those are remotly stored and which
            are not present within the local QDT plugins folder.

        Args:
            input_list (List[QgisPlugin]): list of plugins to filter

        Returns:
            List[QgisPlugin]: list of plugins to download
        """
        plugins_to_download = []

        for plugin in input_list:
            # keep only if remote
            if plugin.location != "remote":
                logger.debug(
                    f"Ignoring plugin '{plugin.name}' because it's not stored remotly."
                )
                continue

            # build destination path
            plugin_download_path = Path(
                self.qdt_plugins_folder, f"{plugin.id_with_version}.zip"
            )

            # check if file already exists
            if plugin_download_path.is_file():
                logger.debug(
                    f"Plugin already exists at {plugin_download_path}, so it "
                    "won't be downloaded."
                )
                continue

            plugins_to_download.append(plugin)

        return plugins_to_download

    def filter_list_copiable_plugins(
        self, input_list: list[QgisPlugin]
    ) -> list[QgisPlugin]:
        """Filter input list of plugins keeping only those which are stored locally and
            which are not present within the local QDT plugins folder.

        Args:
            input_list (List[QgisPlugin]): list of plugins to filter

        Returns:
            List[QgisPlugin]: list of plugins to copy from local disk or network
        """
        plugins_to_copy = []

        for plugin in input_list:
            # keep only if remote
            if plugin.location != "local":
                logger.debug(
                    f"Ignoring plugin '{plugin.name}' because it's not stored locally."
                )
                continue

            # build destination path
            plugin_destination_path = Path(
                self.qdt_plugins_folder, f"{plugin.id_with_version}.zip"
            )

            # check if file already exists
            if plugin_destination_path.is_file():
                logger.debug(
                    f"Plugin already exists at {plugin_destination_path}, so it "
                    "won't be copied from source."
                )
                continue

            plugins_to_copy.append(plugin)

        return plugins_to_copy


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_slugifier
        # for specific test
        python -m unittest tests.test_utils_slugifier.TestUtilsSlugify.test_slugger
"""

# standard library
import unittest

# project
from qgis_deployment_toolbelt.utils.slugger import sluggy

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsSlugify(unittest.TestCase):
    """Test package utilities."""

    def test_slugger(self):
        """Test minimalist slugify function."""
        # hyphen by default
        self.assertEqual(
            sluggy("Oyé oyé brâves gens de 1973 ! Hé oh ! Sentons-nous l'ail %$*§ ?!"),
            "oye-oye-braves-gens-de-1973-he-oh-sentons-nous-lail",
        )

        # with underscore
        self.assertEqual(
            sluggy("Nín hǎo. Wǒ shì zhōng guó rén", "_"),
            "nin_hao_wo_shi_zhong_guo_ren",
        )


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_utils_proxies
        # for specific test
        python -m unittest tests.test_utils_proxies.TestUtilsNetworkProxies.test_proxy_settings
"""

# standard library
import unittest
from os import environ

# project
from qgis_deployment_toolbelt.utils.proxies import get_proxy_settings

# ############################################################################
# ########## Classes #############
# ################################


class TestUtilsNetworkProxies(unittest.TestCase):
    """Test network proxy utilities."""

    def test_proxy_settings(self):
        """Test proxy settings retriever."""
        # by default, no proxy
        self.assertDictEqual(get_proxy_settings(), {})

        # using generic - only http
        get_proxy_settings.cache_clear()
        environ["HTTP_PROXY"] = "http://proxy.example.com:3128"
        self.assertIsInstance(get_proxy_settings(), dict)
        self.assertEqual(
            get_proxy_settings().get("http"), "http://proxy.example.com:3128"
        )
        self.assertIsNone(get_proxy_settings().get("https"))

        environ.pop("HTTP_PROXY")  # clean up

        # using generic - only https
        get_proxy_settings.cache_clear()
        environ["HTTPS_PROXY"] = "https://proxy.example.com:3128"
        self.assertIsInstance(get_proxy_settings(), dict)
        self.assertEqual(
            get_proxy_settings().get("https"), "https://proxy.example.com:3128"
        )
        self.assertIsNone(get_proxy_settings().get("http"))

        environ.pop("HTTPS_PROXY")  # clean up

        # using generic - both http and https
        get_proxy_settings.cache_clear()
        environ["HTTP_PROXY"] = "http://proxy.example.com:3128"
        environ["HTTPS_PROXY"] = "https://proxy.example.com:3128"
        self.assertIsInstance(get_proxy_settings(), dict)
        self.assertEqual(
            get_proxy_settings().get("https"), "https://proxy.example.com:3128"
        )
        self.assertEqual(
            get_proxy_settings().get("http"), "http://proxy.example.com:3128"
        )

        environ.pop("HTTP_PROXY")  # clean up
        environ.pop("HTTPS_PROXY")  # clean up

        # using custom QDT
        get_proxy_settings.cache_clear()
        environ["QDT_PROXY_HTTP"] = "http://user:p8ùX45@proxy.example.com:3128"
        self.assertIsInstance(get_proxy_settings(), dict)
        self.assertEqual(
            get_proxy_settings().get("http"),
            "http://user:p8ùX45@proxy.example.com:3128",
        )
        self.assertEqual(
            get_proxy_settings().get("https"),
            "http://user:p8ùX45@proxy.example.com:3128",
        )

        environ.pop("QDT_PROXY_HTTP")  # clean up

        # not valid URL - just to check the case
        environ["QDT_PROXY_HTTP"] = "socks5://user:motdepasse@proxy.example.com:1182"
        environ.pop("QDT_PROXY_HTTP")  # clean up
        get_proxy_settings.cache_clear()


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa E265

"""Usage from the repo root folder:

    .. code-block:: python

        # for whole test
        python -m unittest tests.job_qgis_installation_finder
        # for specific
        python -m unittest tests.job_qgis_installation_finder
            .TestJobQgisInstallationFinder.test_get_latest_version_from_list
"""

# #############################################################################
# ########## Libraries #############
# ##################################


# Standard library
import unittest

# package
from qgis_deployment_toolbelt.jobs.job_qgis_installation_finder import (
    JobQgisInstallationFinder,
)

# 3rd party


# #############################################################################
# ########## Classes ###############
# ##################################


class TestJobQgisInstallationFinder(unittest.TestCase):
    """Test module."""

    # -- Standard methods --------------------------------------------------------
    @classmethod
    def setUpClass(cls):
        """Executed when module is loaded before any test."""

    # standard methods
    def setUp(self):
        """Fixtures prepared before each test."""
        pass

    def tearDown(self):
        """Executed after each test."""
        pass

    # -- TESTS ---------------------------------------------------------
    def test_get_latest_version_from_list(self):
        """Test definition of latest version from a list of version"""
        self.assertEqual(
            JobQgisInstallationFinder._get_latest_version_from_list(
                ["3.25.1", "4.0.1", "3.28.2"]
            ),
            "4.0.1",
        )

    def test_get_latest_matching_version_path(self):
        """Test definition of latest version from a list of version"""

        # Matching version
        self.assertEqual(
            JobQgisInstallationFinder._get_latest_matching_version_path(
                {
                    "3.25.1": "/path/to/3_25_1",
                    "3.25.8": "/path/to/3_25_8",
                    "3.25.2": "/path/to/3_25_2",
                },
                "3.25",
            ),
            "/path/to/3_25_8",
        )

        # No matching version
        self.assertIsNone(
            JobQgisInstallationFinder._get_latest_matching_version_path(
                {
                    "3.25.1": "/path/to/3_25_1",
                    "3.25.8": "/path/to/3_25_8",
                    "3.25.2": "/path/to/3_25_2",
                },
                "3.36",
            )
        )



#! python3  # noqa E265

"""
    Usage from the repo root folder:

    .. code-block:: bash
        # for whole tests
        python -m unittest tests.test_shortcuts
        # for specific test
        python -m unittest tests.test_shortcuts.TestShortcut.test_shortcut_creation
"""

# standard lib
import unittest
from pathlib import Path
from sys import platform as opersys

# project
from qgis_deployment_toolbelt.shortcuts.shortcuts_handler import ApplicationShortcut

# ############################################################################
# ########## Classes #############
# ################################


class TestShortcut(unittest.TestCase):
    """Test shortcut creation and deletion."""

    def test_shortcut_creation_complete(self):
        """Test creation of shortcut."""
        shortcut = ApplicationShortcut(
            name="QDT - Test shortcut complete",
            exec_path=Path(__file__),
            exec_arguments=("--test", "unit"),
            description="A test shortcut",
            icon_path=Path(__file__).parent / "icon.png",
            work_dir=Path(__file__).parent,
        )

        shortcuts_paths = shortcut.create(
            desktop=True,
            start_menu=True,
        )

        if opersys in ("linux", "win32"):
            self.assertIsInstance(shortcuts_paths, tuple)
            self.assertIsInstance(shortcuts_paths[0], Path)
            self.assertIsInstance(shortcuts_paths[1], Path)

            self.assertTrue(shortcuts_paths[0].exists())
            self.assertTrue(shortcuts_paths[1].exists())

            # clean up
            shortcuts_paths[0].unlink()
            shortcuts_paths[1].unlink()

    def test_shortcut_creation_minimal(self):
        """Test creation of shortcut."""
        shortcut = ApplicationShortcut(
            name="QDT - Test shortcut minimal",
            exec_path=Path(__file__),
            work_dir=Path(__file__).parent,
        )

        shortcuts_paths = shortcut.create(desktop=True, start_menu=True)

        if opersys in ("linux", "win32"):
            self.assertIsInstance(shortcuts_paths, tuple)
            self.assertIsInstance(shortcuts_paths[0], Path)
            self.assertIsInstance(shortcuts_paths[1], Path)

            self.assertTrue(shortcuts_paths[0].exists())
            self.assertTrue(shortcuts_paths[1].exists())

            # clean up
            shortcuts_paths[0].unlink()
            shortcuts_paths[1].unlink()

    def test_shortcut_bad_types(self):
        """Test shortcut TypeError raises."""
        with self.assertRaises(TypeError):
            ApplicationShortcut(
                name=["QDT - Test shortcut bad type"],
                exec_path=Path(__file__),
            )

        with self.assertRaises(TypeError):
            ApplicationShortcut(
                name="QDT - Test shortcut bad type",
                exec_path=[Path(__file__)],
            )

        with self.assertRaises(TypeError):
            ApplicationShortcut(
                name="QDT - Test shortcut bad type",
                exec_path=Path(__file__),
                exec_arguments="--test unit",
            )

        with self.assertRaises(TypeError):
            ApplicationShortcut(
                name="QDT - Test shortcut bad type",
                exec_path=Path(__file__),
                exec_arguments=("--test", "unit"),
                description=["A test shortcut"],
            )

        with self.assertRaises(TypeError):
            ApplicationShortcut(
                name="QDT - Test shortcut bad type",
                exec_path=Path(__file__),
                exec_arguments=("--test", "unit"),
                description="A test shortcut",
                icon_path=["icon.png"],
            )

        with self.assertRaises(TypeError):
            ApplicationShortcut(
                name="QDT - Test shortcut bad type",
                exec_path=Path(__file__),
                work_dir=["C:\\"],
            )


# ############################################################################
# ####### Stand-alone run ########
# ################################
if __name__ == "__main__":
    unittest.main()



#! python3  # noqa: E265

"""
    Read and validate scenario files.

    Author: Julien Moura (https://github.com/guts, Oslandia)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os import environ

# project
from qgis_deployment_toolbelt.jobs.generic_job import GenericJob
from qgis_deployment_toolbelt.jobs.job_environment_variables import (
    JobEnvironmentVariables,
)
from qgis_deployment_toolbelt.jobs.job_plugins_downloader import JobPluginsDownloader
from qgis_deployment_toolbelt.jobs.job_plugins_synchronizer import (
    JobPluginsSynchronizer,
)
from qgis_deployment_toolbelt.jobs.job_profiles_downloader import JobProfilesDownloader
from qgis_deployment_toolbelt.jobs.job_profiles_synchronizer import (
    JobProfilesSynchronizer,
)
from qgis_deployment_toolbelt.jobs.job_qgis_installation_finder import (
    JobQgisInstallationFinder,
)
from qgis_deployment_toolbelt.jobs.job_shortcuts import JobShortcutsManager
from qgis_deployment_toolbelt.jobs.job_splash_screen import JobSplashScreenManager

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Classes ###############
# ##################################


class JobsOrchestrator:
    """Orchestrate jobs."""

    JOBS: tuple = (
        JobEnvironmentVariables,
        JobPluginsDownloader,
        JobPluginsSynchronizer,
        JobProfilesDownloader,
        JobProfilesSynchronizer,
        JobShortcutsManager,
        JobSplashScreenManager,
        JobQgisInstallationFinder,
    )
    PACKAGE_NAME: str = "qgis_deployment_toolbelt.jobs"

    def __init__(self) -> None:
        """Instanciate orchestrator."""
        # log environment variables prefixed with QDT_
        qdt_env_vars = {
            env_var: value
            for env_var, value in environ.items()
            if env_var.startswith("QDT_")
        }
        if nb_qdt_envvars := len(qdt_env_vars):
            logger.debug(f"{nb_qdt_envvars} environment variables related to QDT:")
            for var_name, var_value in qdt_env_vars.items():
                logger.debug(f"{var_name}={var_value}")

    @property
    def jobs_ids(self) -> tuple[str, ...]:
        """Returns available jobs ID.

        Returns:
            tuple[str]: tuple of jobs ids
        """
        return tuple([job.ID for job in self.JOBS])

    def get_job_module_from_id(self, job_id: str) -> GenericJob | None:
        """Get job class from id.

        :param str job_id: job identifier (as defined in scenario file)
        :return object: Job class
        """
        for job in self.JOBS:
            if job.ID == job_id:
                return job

    def init_job_class_from_id(self, job_id: str, options: dict) -> GenericJob | None:
        """Get job class from id and instanciate it with options.

        :param str job_id: job identifier (i.e. "qprofiles-manager")
        :param dict options: job options

        :return object: instanciated job class with options
        """
        if job := self.get_job_module_from_id(job_id):
            return job(options)
